import threading
from .common import *
from .query import *
from .eng_taosrestful import taos_resetful
class Database(object):
    commit_select = False
    compiler_class = QueryCompiler
    empty_limit = False
    field_overrides = {}
    for_update = False
    interpolation = '{}'
    op_overrides = {}
    quote_char = '"'
    reserved_tables = []
    sequences = False
    subquery_delete_same_table = True

    def __init__(self, database,canDict=False, threadlocals=False, autocommit=True,
                 fields=None, ops=None, **connect_kwargs):
        self.init(database, **connect_kwargs)

        if threadlocals:
            self.__local = threading.local()
        else:
            self.__local = type('DummyLocal', (object,), {})

        self._conn_lock = threading.Lock()
        self.autocommit = autocommit
        self.databases = {}
        self.tables = {}
        self.stables = {}
        self.field_overrides = dict_update(self.field_overrides, fields or {})
        self.op_overrides = dict_update(self.op_overrides, ops or {})
        self.curSql =''
        self.canDict = canDict

    def init(self, database, **connect_kwargs):
        self.deferred = database is None
        self.database = database
        self.connect_kwargs = connect_kwargs

    def connect(self):
        with self._conn_lock:
            if self.deferred:
                raise Exception('Error, database not properly initialized before opening connection')
            self.__local.conn = self._connect(self.database, **self.connect_kwargs)
            self.__local.closed = False
            self.create_database(safe=True)
            # self.get_databases()

    def close(self):
        with self._conn_lock:
            if self.deferred:
                raise Exception('Error, database not properly initialized before closing connection')
            self._close(self.__local.conn)
            self.__local.closed = True

    def get_conn(self):
        if not hasattr(self.__local, 'closed') or self.__local.closed:
            self.connect()
        return self.__local.conn

    def is_closed(self):
        return getattr(self.__local, 'closed', True)

    def get_cursor(self):
        return self.get_conn().cursor()

    def _close(self, conn):
        conn.close()

    def _connect(self, database, **kwargs):
        raise NotImplementedError

    @classmethod
    def register_fields(cls, fields):
        cls.field_overrides = dict_update(cls.field_overrides, fields)

    @classmethod
    def register_ops(cls, ops):
        cls.op_overrides = dict_update(cls.op_overrides, ops)

    def rows_affected(self, cursor):
        return cursor.rowcount

    def get_compiler(self):
        return self.compiler_class(
            self.quote_char, self.interpolation, self.field_overrides,
            self.op_overrides)

    def execute(self, query):
        sql, params = query.sql(self.get_compiler())
        if isinstance(query, SelectQuery):
            commit = self.commit_select
        else:
            commit = query.require_commit
        return self.execute_sql(sql, params, commit)

    def execute_sql(self, sql, params=None, require_commit=True):
        cursor = self.get_cursor()
        # logger.debug((sql, params))
        res = cursor.execute(sql, params or (),self.canDict)
        self.curSql =cursor.sqlStr
        if res:
            if require_commit:
                self.commit()
            # logger.debug(cursor.head,str(cursor.rows))
            return cursor
        return None

    def commit(self):
        self.get_conn().commit()

    def rollback(self):
        self.get_conn().rollback()
    
    def get_supertables(self):
        qc = self.get_compiler()
        res= self.execute_sql(qc.show_tables(self.database,super=True))
        self.stables={}
        for item in res.data:
            if item:
                self.stables[item[0]] = dict(zip(res.head, item))
        return self.stables

    def get_tables(self):
        qc = self.get_compiler()
        res= self.execute_sql(qc.show_tables(self.database,super=False))
        self.tables={}
        for item in res.data:
            if item:
                self.tables[item[0]] = dict(zip(res.head, item))
        return self.tables

    # def get_indexes_for_table(self, table):
    #     raise NotImplementedError

    def create_database(self,safe=False,keep= None,comp=None,replica=None,quorum=None,blocks=None):
        qc = self.get_compiler()
        res = self.execute_sql(qc.create_database(self.database,safe=safe,keep= keep,comp=comp,replica=replica,quorum=quorum,blocks=blocks))
        self.get_databases()
        return [[0]] == res
    
    def get_databases(self):
        qc = self.get_compiler()
        res= self.execute_sql(qc.show_database(self.database))
        self.databases={}
        for item in res.data:
            if item:
                self.databases[item[0]] = dict(zip(res.head, item))
        return self.databases
    
    def drop_database(self,safe=True):
        qc = self.get_compiler()
        res = self.execute_sql(qc.drop_database(self.database,safe=safe))
        self.get_databases()
        return [[0]] == res
    
    def alter_database(self,keep= None,comp=None,replica=None,quorum=None,blocks=None):
        qc = self.get_compiler()
        res = self.execute_sql(qc.alter_database(self.database,keep= keep,comp=comp,replica=replica,quorum=quorum,blocks=blocks))
        self.get_databases()
        return [[0]] == res

    def create_table(self, model_class,safe=False):
        qc = self.get_compiler()
        return [[0]] == self.execute_sql(qc.create_table(model_class,safe=safe))

    def drop_table(self, model_class, safe=False):
        qc = self.get_compiler()
        return [[0]] == self.execute_sql(qc.drop_table(model_class, safe))
    
    def describe_table(self, model_class):
        qc = self.get_compiler()
        return self.execute_sql(qc.describe_table(model_class))
    def describe_table_name(self, table_name):
        return self.execute_sql('DESCRIBE %s.%s' % (self.database,table_name),[])
    def change_table_tagvalue(self, model_class,name,value):
        qc = self.get_compiler()
        return self.execute_sql(qc.change_table_tagvalue(model_class,name,value))
    def add_tag(self, model_class,value):
        qc = self.get_compiler()
        return self.execute_sql(qc.add_tag(model_class,value))
    def drop_tag(self, model_class,name):
        qc = self.get_compiler()
        return self.execute_sql(qc.drop_tag(model_class,name))
    def change_tag_name(self, model_class,name,newname):
        qc = self.get_compiler()
        return self.execute_sql(qc.change_tag_name(model_class,name,newname))
    def raw_sql(self, sql, *params):
        return self.execute_sql(sql.replace("?", "{}"),params)
class TdEngineDatabase(Database):
    def _connect(self, database, **kwargs):
        return taos_resetful.connect(database=database, **kwargs)
default_database = TdEngineDatabase('demo',host='localhost')


import datetime
import re
from .common import *

class Leaf(object):
    def __init__(self):
        self.negated = False
        self._alias = None

    def __invert__(self):
        self.negated = not self.negated
        return self

    def alias(self, a):
        self._alias = a
        return self

    def asc(self):
        return Ordering(self, True)

    def desc(self):
        return Ordering(self, False)

    def _e(op, inv=False):
        def inner(self, rhs):
            if inv:
                return Expr(rhs, op, self)
            return Expr(self, op, rhs)
        return inner
    __and__ = _e(OP_AND)
    __or__ = _e(OP_OR)

    __add__ = _e(OP_ADD)
    __sub__ = _e(OP_SUB)
    __mul__ = _e(OP_MUL)
    __truediv__ = _e(OP_DIV)
    __xor__ = _e(OP_XOR)
    __radd__ = _e(OP_ADD, inv=True)
    __rsub__ = _e(OP_SUB, inv=True)
    __rmul__ = _e(OP_MUL, inv=True)
    __rtruediv__ = _e(OP_DIV, inv=True)
    __rand__ = _e(OP_AND, inv=True)
    __ror__ = _e(OP_OR, inv=True)
    __rxor__ = _e(OP_XOR, inv=True)

    __eq__ = _e(OP_EQ)
    __lt__ = _e(OP_LT)
    __le__ = _e(OP_LTE)
    __gt__ = _e(OP_GT)
    __ge__ = _e(OP_GTE)
    __ne__ = _e(OP_NE)
    __lshift__ = _e(OP_IN)
    __rshift__ = _e(OP_IS)
    __mod__ = _e(OP_LIKE)
    __pow__ = _e(OP_ILIKE)

class Expr(Leaf):
    def __init__(self, lhs, op, rhs, negated=False):
        super(Expr, self).__init__()
        self.lhs = lhs
        self.op = op
        self.rhs = rhs
        self.negated = negated

    def clone(self):
        return Expr(self.lhs, self.op, self.rhs, self.negated)

class DQ(Leaf):
    def __init__(self, **query):
        super(DQ, self).__init__()
        self.query = query

    def clone(self):
        return DQ(**self.query)

class Param(Leaf):
    def __init__(self, data):
        self.data = data
        super(Param, self).__init__()

class Func(Leaf):
    def __init__(self, name, *params):
        self.name = name
        self.params = params
        super(Func, self).__init__()

    def clone(self):
        return Func(self.name, *self.params)

    def __getattr__(self, attr):
        def dec(*args, **kwargs):
            return Func(attr, *args, **kwargs)
        return dec

fn = Func(None)

class FieldDescriptor(object):
    def __init__(self, field):
        self.field = field
        self.att_name = self.field.name
        self.column_name = self.field.db_column

    def __get__(self, instance, instance_type=None):
        if instance:
            return instance._data.get(self.att_name)
        return self.field

    def __set__(self, instance, value):
        instance._data[self.att_name] = value

class Field(Leaf):
    _field_counter = 0
    _order = 0
    db_field = 'unknown'
    template = '%(column_type)s'

    def __init__(self, null=False, index=False,  verbose_name=None, #unique=False,primary_key=False, sequence=None,choices=None,
                 help_text=None, db_column=None, default=None, *args, **kwargs):
        self.null = null
        self.index = index
        # self.unique = unique
        self.verbose_name = verbose_name
        self.help_text = help_text
        self.db_column = db_column
        self.default = default
        # self.choices = choices
        # self.primary_key = primary_key
        # self.sequence = sequence

        self.attributes = self.field_attributes()
        self.attributes.update(kwargs)

        Field._field_counter += 1
        self._order = Field._field_counter

        super(Field, self).__init__()
    def __str__(self) -> str:
        return "null: %s\nindex: %s\nverbose_name: %s\ndb_column: %s\ndefault: %s\n" % \
            (self.null,self.index,self.verbose_name,self.db_column,self.default)

    def add_to_class(self, model_class, name):
        self.name = name
        self.model_class = model_class
        self.db_column = self.db_column or self.name
        self.verbose_name = self.verbose_name or re.sub('_+', ' ', name).title()

        model_class._meta.fields[self.name] = self
        model_class._meta.columns[self.db_column] = self
        setattr(model_class, name, FieldDescriptor(self))

    def add_tags_to_class(self, model_class, name):
        self.name = name
        self.model_class = model_class
        self.db_column = self.db_column or self.name
        self.verbose_name = self.verbose_name or re.sub('_+', ' ', name).title()

        model_class._tags.fields[self.name] = self
        model_class._tags.columns[self.db_column] = self
        setattr(model_class, name, FieldDescriptor(self))
    def field_attributes(self):
        return {}

    def get_db_field(self):
        return self.db_field

    def coerce(self, value):
        return value

    def db_value(self, value):
        return 'null' if value is None else self.coerce(value)
    def python_value(self, value):
        if value == 'null':
            return None
        return value if value is None else self.coerce(value)
    def count(self):
        return fn.COUNT(self)
    def avg(self):
        return fn.AVG(self)
    def twa(self):
        return fn.TWA(self)
    def sum(self):
        return fn.SUM(self)
    def stddev(self):
        return fn.STDDEV(self)
    def min(self):
        return fn.MIN(self)
    def max(self):
        return fn.MAX(self)
    def last(self):
        return fn.LAST(self)
    def first(self):
        return fn.FIRST(self)
    def last_row(self):
        return fn.LAST_ROW(self)
    def spread(self):
        return fn.SPREAD(self)
    def diff(self):
        return fn.DIFF(self)
    def top(self,top=1):
        return fn.TOP(self,top)
    def bottom(self,bottom=1):
        return fn.BOTTOM(self,bottom)
    def apercentile(self,p=1):
        return fn.APERCENTILE(self,p)
    def percentile(self,p=1):
        return fn.PERCENTILE(self,p)
    def leastsquares(self,start_val,step_val):
        return fn.LEASTSQUARES(self,start_val,step_val)
class IntegerField(Field):
    db_field = 'int'

    def coerce(self, value):
        return int(value)
class BigIntegerField(IntegerField):
    db_field = 'bigint'
class SmallIntegerField(IntegerField):
    db_field = 'smallint'
class TinyIntegerField(IntegerField):
    db_field = 'tinyint'
class FloatField(Field):
    db_field = 'float'

    def coerce(self, value):
        return float(value)
class DoubleField(FloatField):
    db_field = 'double'
class NCharField(Field):
    db_field = 'nchar'
    template = '%(column_type)s(%(max_length)s)'

    def field_attributes(self):
        return {'max_length': 255}

    def coerce(self, value):
        value = value or ''
        return value[:self.attributes['max_length']]
class BinaryField(Field):
    db_field = 'binary'
    template = '%(column_type)s(%(max_length)s)'

    def field_attributes(self):
        return {'max_length': 255}

    def coerce(self, value):
        value =  value or ''
        return value[:self.attributes['max_length']]
def format_date_time(value, formats, post_process=None):
    post_process = post_process or (lambda x: x)
    for fmt in formats:
        try:
            return post_process(datetime.datetime.strptime(value, fmt))
        except ValueError:
            pass
    return value
class DateTimeField(Field):
    db_field = 'datetime'

    def field_attributes(self):
        return {
            'formats': [
                '%Y-%m-%d %H:%M:%S.%f',
                '%Y-%m-%d %H:%M:%S',
                '%Y-%m-%d',
            ]
        }

    def python_value(self, value):
        if value and isinstance(value, str):
            return format_date_time(value, self.attributes['formats'])
        return value
class PrimaryKeyField(DateTimeField):
    db_field = 'primary_key'
class BooleanField(Field):
    db_field = 'bool'

    def coerce(self, value):
        return bool(value)



import logging
import requests
from datetime import datetime
from .common import logger

# logging.basicConfig(level=logging.DEBUG)
# logger = logging.getLogger('crown')

PATH='rest/sql'

class Row(list):
    def __init__(self,arr,head):
        super(Row, self).__init__(arr)
        self.head = head
    def __getitem__(self, key):
        try:
            if isinstance(key,int):
                return super(Row, self).__getitem__(key)
            else:
                return super(Row, self).__getitem__(self.head.index(key))
        except:
            return None
    def get(self, key):
        try:
            if isinstance(key,int):
                return super(Row, self).__getitem__(key)
            else:
                return super(Row, self).__getitem__(self.head.index(key))
        except:
            return None
        
class Cursor(list):
    def __init__(self,conn):
        super(Cursor, self).__init__([])
        self.conn = conn
        self.status=''
        self.rowcount = 0
        self.head = []
        self.data = []
        self.err_code = None
        self.err_desc = ''
        self.sqlStr =''
    def execute(self,sql,param=(),canDict=False):
        if param:
            param = map(lambda x: '"%s"' % x if isinstance(x,str) or isinstance(x,datetime) else x, param)
            sql = sql.format(*param)
        logger.debug(("excute sql->",sql))
        self.sqlStr =sql
        res = self.conn.execute_sql(sql)
        logger.debug(("raw result->",res))
        if res:
            self.status=res.get('status')
            if self.status == 'succ':
                self.rowcount = res.get('rows')
                self.head = res.get('head')
                if canDict:
                    for d in res.get('data') or []:
                        self.data.append(Row(d,self.head))
                else:
                    self.data = res.get('data')
                logger.debug(("result->",res.get('data'), res.get('head')))
                super(Cursor, self).__init__(self.data)
                return True
            else:
                self.err_code = res.get('code')
                self.err_desc = res.get('desc')
                #自动建立数据库
                if self.err_code == 896:
                    res = self.execute('create database %s' % self.conn.database)
                    if res:
                        return self.execute(sql)
                raise Exception(self.err_desc)
        else:
            raise Exception('server error')


class Conn():
    def __init__(self,database,url,user,passwd):
        self.user = user
        self.passwd = passwd
        self.url = url
        self.database = database

    def execute_sql(self,sql):
        reponse = requests.post(self.url, auth=(self.user, self.passwd),data=sql.encode())
        # if reponse.status_code == requests.codes.ok:
        data = reponse.json()
        return data
    
    def cursor(self):
        cursor = Cursor(self)
        return cursor
    
    def commit(self):
        pass
    def rollback(self):
        pass
class TaosRestful():
    
    def connect(self,database, host='localhost',port=6041,user='root',passwd='taosdata'):
        conn = Conn(database,'http://%s:%s/%s' % (host,port,PATH),user,passwd)
        return conn
        
taos_resetful = TaosRestful()

       


from copy import deepcopy
from .field import *
from .database import *
from .query import *
import re
from datetime import datetime
# __all__ = [
#     'IntegerField', 'BigIntegerField', 'PrimaryKeyField', 'FloatField', 'DoubleField',
#     'DateTimeField','BooleanField', 'Model', 'DoesNotExist',  'Field','SmallIntegerField',
#     'TinyIntegerField','NCharField','BinaryField','TdEngineDatabase','SuperModel'
# ]
class ModelOptions(object):
    def __init__(self, cls, database=None, db_table=None,
                 order_by=None, primary_key=None):
        self.model_class = cls
        self.name = cls.__name__.lower()
        self.fields = {}
        self.columns = {}
        self.defaults = {}
        self.super_table = False
        self.child_table = False
        self.my_template =''

        self.database = database or default_database
        self.db_table = db_table
        self.order_by = order_by
        self.primary_key_name = primary_key
        self.primary_key = None
        # self.rel = {}
        # self.reverse_rel = {}
    def __str__(self) -> str:
        return "name: %s\ndatabase: %s\norder_by: %s\nprimary_key: %s\ndb_table: %s\nfields: %s\ncolumns: %s\ndefaults: %s\n" % \
            (self.name,self.database,self.order_by,self.primary_key,self.db_table,self.fields,self.columns,self.defaults)
    def prepared(self):
        for field in self.fields.values():
            if field.default is not None:
                self.defaults[field] = field.default

        if self.order_by:
            norm_order_by = []
            for clause in self.order_by:
                if isinstance(clause,str):
                    field = self.fields[clause.lstrip('-')]
                    if clause.startswith('-'):
                        norm_order_by.append(field.desc())
                    else:
                        norm_order_by.append(field.asc())
                else:
                    norm_order_by.append(clause)
            self.order_by = norm_order_by

    def get_default_dict(self):
        dd = {}
        for field, default in self.defaults.items():
            if callable(default):
                dd[field.name] = default()
            else:
                dd[field.name] = default
        return dd

    def get_sorted_fields(self):
        return sorted(self.fields.items(), key=lambda v: (isinstance(v[1], PrimaryKeyField) and 1 or 2, v[1]._order))

    def get_field_names(self):
        return [f[0] for f in self.get_sorted_fields()]

    def get_fields(self):
        return [f[1] for f in self.get_sorted_fields()]


class BaseModel(type):
    inheritable_options = ['database', 'order_by']

    def __new__(cls, name, bases, attrs):
        if not bases:
            return super(BaseModel, cls).__new__(cls, name, bases, attrs)

        meta_options = {}
        tags_options = {}
        meta = attrs.pop('Meta', None)
        if meta:
            meta_options.update((k, v) for k, v in meta.__dict__.items() if not k.startswith('_') and not isinstance(v, Field))
            tags_options.update((k, v) for k, v in meta.__dict__.items() if not k.startswith('_') \
                                                                            and isinstance(v, Field) \
                                                                            and not isinstance(v, PrimaryKeyField) \
                                                                            and not isinstance(v, DateTimeField) \
                                                                            and not isinstance(v, NCharField))

        for b in bases:
            if not hasattr(b, '_meta'):
                continue
            base_meta = getattr(b, '_meta')
            for (k, v) in base_meta.__dict__.items():
                if k in cls.inheritable_options and k not in meta_options:
                    meta_options[k] = v

            for (k, v) in b.__dict__.items():
                if isinstance(v, FieldDescriptor) and k not in attrs:
                    if not isinstance(v.field, PrimaryKeyField):
                        attrs[k] = deepcopy(v.field)

        cls = super(BaseModel, cls).__new__(cls, name, bases, attrs)
        cls._meta = ModelOptions(cls, **meta_options)
        cls._tags = ModelOptions(cls) if tags_options else None
        for name, attr in tags_options.items():
            if isinstance(attr, Field):
                attr.add_tags_to_class(cls, name)
        cls._data = None

        primary_key = None
        for name, attr in cls.__dict__.items():
            if isinstance(attr, Field):
                attr.add_to_class(cls, name)
                if isinstance(attr, PrimaryKeyField):
                    primary_key = attr
        if not primary_key:
            primary_key = PrimaryKeyField()
            primary_key.add_to_class(cls, cls._meta.primary_key_name or 'ts')

        cls._meta.primary_key = primary_key

        if not cls._meta.db_table:
            cls._meta.db_table = re.sub('[^\\w]+', '_', cls.__name__.lower())
        
        cls._meta.db_table = "%s.%s" % (cls._meta.database.database,cls._meta.db_table)

        # create a repr and error class before finalizing
        if hasattr(cls, '__unicode__'):
            setattr(cls, '__repr__', lambda self: '<%s: %r>' % (
                cls.__name__, self.__unicode__()))

        exception_class = type('%sDoesNotExist' % cls.__name__, (DoesNotExist,), {})
        cls.DoesNotExist = exception_class
        cls._meta.prepared()
        if cls._tags:
            cls._tags.prepared()

        return cls


class Model(metaclass=BaseModel):
    # __metaclass__ = BaseModel

    def __init__(self, *args, **kwargs):
        self._data = self._meta.get_default_dict()
        self._obj_cache = {} # cache of related objects

        for k, v in kwargs.items():
            setattr(self, k, v)
    def get_ts(self):
        return getattr(self, self._meta.primary_key.name)
    def set_ts(self, ts):
        setattr(self, self._meta.primary_key.name, ts)
    @classmethod
    def create_table(cls,safe=True):
        return cls._meta.database.create_table(cls,safe)
    @classmethod
    def dynamic_create_table(cls,table_name,database=default_database,safe=True,**fields):
        fields['Meta']= type('Meta', (object,), dict(database = database,db_table=table_name))
        resModel = type(table_name, (cls,), fields)
        resModel.create_table(safe=safe)
        return resModel
    @classmethod
    def model_from_table(cls,table_name,database=default_database):
        if table_name in database.get_tables():
            fields = {}
            fields['Meta']= type('Meta', (object,), dict(database = database,db_table=table_name))
            desc_fields = database.describe_table_name(table_name)
            for i,f in enumerate(desc_fields):
                if f[3] != 'TAG':
                    val = None
                    if f[1] == 'TIMESTAMP':
                        val = PrimaryKeyField(column_name=f[0]) if i == 0 else DateTimeField(column_name=f[0])
                    elif f[1] == 'FLOAT':
                        val = FloatField(column_name=f[0])
                    elif f[1] == 'DOUBLE':
                        val = DoubleField(column_name=f[0])
                    elif f[1] == 'INT':
                        val = IntegerField(column_name=f[0])
                    elif f[1] == 'BIGINT':
                        val = BigIntegerField(column_name=f[0])
                    elif f[1] == 'SMALLINT':
                        val = SmallIntegerField(column_name=f[0])
                    elif f[1] == 'TINYINT':
                        val = TinyIntegerField(column_name=f[0])
                    elif f[1] == 'NCHAR':
                        val = NCharField(column_name=f[0],max_length=f[2])
                    elif f[1] == 'BINARY':
                        val = BinaryField(column_name=f[0],max_length=f[2])
                    elif f[1] == 'BOOL':
                        val = BooleanField(column_name=f[0])
                    else:
                        raise Exception('have unknow field')
                    fields[f[0]] = val
            resModel = type(table_name, (cls,), fields)
            return resModel
        else:
            return None
        
    @classmethod
    def drop_table(cls,safe=True):
        return cls._meta.database.drop_table(cls,safe)
    @classmethod
    def change_tag_value(cls,**kwargs):
        res = None
        for name,value in kwargs.items():
            res = cls._meta.database.change_table_tagvalue(cls,name,value)
            if res == None:
                return res
        return res
    @classmethod
    def describe_table(cls):
        return cls._meta.database.describe_table(cls)
    @classmethod
    def table_exists(cls):
        return cls._meta.db_table[cls._meta.db_table.index('.')+1:] in cls._meta.database.get_tables() or \
                 cls._meta.db_table[cls._meta.db_table.index('.')+1:].lower() in cls._meta.database.get_tables()
        # return cls._meta.db_table in ["%s.%s" % (cls._meta.database.database,x[0]) for x in cls._meta.database.get_tables()]
    @classmethod
    def select(cls, *selection):
        query = SelectQuery(cls, *selection)
        if cls._meta.order_by:
            query = query.order_by(*cls._meta.order_by)
        return query
    @classmethod
    def insert(cls, **insert):
        pk = cls._meta.primary_key
        if pk.name not in insert:
            now = datetime.now()
            insert[pk.name] = now
        fdict = []
        for field in cls._meta.get_field_names():
            if field in insert:
                value = insert.get(field)
                if isinstance(value,datetime):
                    value = value.strftime("%Y-%m-%d %H:%M:%S.%f")
                fdict.append({'obj':cls._meta.fields[field],'value':value})
        return InsertQuery(cls, fdict).execute()
    def save(self):
        field_dict = dict(self._data)
        pk = self._meta.primary_key
        if pk.name not in field_dict:
            now = datetime.now()
            self.set_ts(now)
            field_dict[pk.name] = now
        # insert = self.insert(**field_dict)
        return self.insert(**field_dict)
    def get(self,expr):
        p1 = re.compile(r'^\((.*?)\)$', re.S)
        query_expr=type(self)._meta.database.get_compiler().parse_expr(expr)
        if query_expr:
            param = map(lambda x: '"%s"' % x if isinstance(x,str) or isinstance(x,datetime) else x, query_expr[1])
            queryori = query_expr[0].format(*param)
            query_bracket = re.findall(p1,queryori)
            if len(query_bracket) >0:
                query_str = query_bracket[0].lower()
            else:
                query_str = queryori.lower()
            if hasattr(self,query_str):
                return getattr(self,query_str)
            elif hasattr(self,'('+query_str+')'):
                return getattr(self,'('+query_str+')')
            else:
                return None
        else:
            return None
    @classmethod
    def fc(cls,col_name):
        for name,field in cls.__dict__.items():
            if isinstance(field,FieldDescriptor) and field.column_name == col_name:
                    return field.field
        return None
    @classmethod
    def f(cls,name):
        return getattr(cls, name, None)
class SuperModel(metaclass=BaseModel):
    def __init__(self, *args, **kwargs):
        self._data = self._meta.get_default_dict()
        self._obj_cache = {} # cache of related objects
        for k, v in kwargs.items():
            setattr(self, k, v)
    @classmethod
    def create_table(cls,safe=True):
        return cls._meta.database.create_table(cls,safe)
    @classmethod
    def dynamic_create_table(cls,table_name,database=default_database,safe=True,tags={},**fields):
        attr = dict(database =  database,db_table=table_name)
        attr.update(tags)
        _meta= type('Meta', (object,), attr)
        fields['Meta'] = _meta
        resModel = type(table_name, (cls,), fields)
        resModel.create_table(safe=safe)
        return resModel
    @classmethod
    def supermodel_from_table(cls,table_name,database=default_database):
        if table_name in database.get_supertables():
            fields = {}
            tags =dict(database = database,db_table=table_name)
            desc_fields = database.describe_table_name(table_name)
            for i,f in enumerate(desc_fields):
                val = None
                if f[1] == 'TIMESTAMP':
                    val = PrimaryKeyField(column_name=f[0]) if i == 0 else DateTimeField(column_name=f[0])
                elif f[1] == 'FLOAT':
                    val = FloatField(column_name=f[0])
                elif f[1] == 'DOUBLE':
                    val = DoubleField(column_name=f[0])
                elif f[1] == 'INT':
                    val = IntegerField(column_name=f[0])
                elif f[1] == 'BIGINT':
                    val = BigIntegerField(column_name=f[0])
                elif f[1] == 'SMALLINT':
                    val = SmallIntegerField(column_name=f[0])
                elif f[1] == 'TINYINT':
                    val = TinyIntegerField(column_name=f[0])
                elif f[1] == 'NCHAR':
                    val = NCharField(column_name=f[0],max_length=f[2])
                elif f[1] == 'BINARY':
                    val = BinaryField(column_name=f[0],max_length=f[2])
                elif f[1] == 'BOOL':
                    val = BooleanField(column_name=f[0])
                else:
                    raise Exception('have unknow field')
                if f[3] == 'TAG':
                    tags[f[0]] = val
                else:
                    fields[f[0]] = val
            fields['Meta']= type('Meta', (object,), tags)
            resModel = type(table_name, (cls,), fields)
            return resModel
        else:
            return None
    @classmethod
    def create_son_table(cls,name,**kwargs):
        cls._meta.database.create_table(cls,safe=True)
        tags =[]
        for field in cls._tags.get_field_names():
            if field not in kwargs:
                raise Exception('tag %s not have' % field)
            tags.append({'obj':cls._tags.fields[field],'value':kwargs.get(field)})
        # fdict = dict((cls._tags.fields[f], v) for f, v in tags.items())
        CreateSonTableQuery(cls, tags,name).execute()
        son_model = type(name, (cls,Model), dict())
        #########################################fix custom primary bug
        primary_column_name = cls._meta.primary_key.db_column
        if primary_column_name != 'ts':
            son_model._meta.primary_key = cls._meta.primary_key
            son_model._meta.fields.pop('ts')
            son_model._meta.columns.pop('ts')
            son_model._meta.fields[primary_column_name] = cls._meta.fields[primary_column_name]
            son_model._meta.columns[primary_column_name] = cls._meta.columns[primary_column_name]
        #############################################
        for name in cls._tags.get_field_names():
            delattr(son_model,name)
            field = son_model._meta.fields[name]
            del son_model._meta.columns[field.db_column]
            del son_model._meta.fields[name]
        son_model._tags =None
        return son_model
    @classmethod
    def drop_table(cls,safe=True):
        return cls._meta.database.drop_table(cls,safe)
    @classmethod
    def describe_table(cls):
        return cls._meta.database.describe_table(cls)
    @classmethod
    def add_tags(cls,*args):
        res = None
        for value in args:
            if isinstance(value, Field) and not isinstance(value, DateTimeField) and value.db_column:
                res = cls._meta.database.add_tag(cls,value)
                if res == None:
                    return res
                else:
                    setattr(cls._meta, value.db_column, value)
        return res
    @classmethod
    def drop_tag(cls,name):
        res = cls._meta.database.drop_tag(cls,name)
        if res != None and hasattr(cls._meta,name):
            delattr(cls._meta,name)
        return res
    @classmethod
    def change_tag_name(cls,name,newname):
        res = cls._meta.database.change_tag_name(cls,name,newname)
        if res != None and hasattr(cls._meta,name):
            tagObj = getattr(cls._meta,name)
            tagObj.db_column=newname
        return res
    @classmethod
    def supertable_exists(cls):
        # tables = ["%s.%s" % (cls._meta.database.database,x[3]) for x in cls._meta.database.get_tables()]
        return cls._meta.db_table[cls._meta.db_table.index('.')+1:] in cls._meta.database.get_supertables() or \
                cls._meta.db_table[cls._meta.db_table.index('.')+1:].lower() in cls._meta.database.get_supertables()
    @classmethod
    def select(cls, *selection):
        query = SelectQuery(cls, *selection)
        if cls._meta.order_by:
            query = query.order_by(*cls._meta.order_by)
        return query
    @classmethod
    def fc(cls,col_name):
        for name,field in cls.__dict__.items():
            if isinstance(field,FieldDescriptor) and field.column_name == col_name:
                    return field.field
        return None
    @classmethod
    def f(cls,name):
        return getattr(cls, name, None)
    def get(self,expr):
        p1 = re.compile(r'^\((.*?)\)$', re.S)
        query_expr=type(self)._meta.database.get_compiler().parse_expr(expr)
        if query_expr:
            param = map(lambda x: '"%s"' % x if isinstance(x,str) or isinstance(x,datetime) else x, query_expr[1])
            queryori = query_expr[0].format(*param)
            query_bracket = re.findall(p1,queryori)
            if len(query_bracket) >0:
                query_str = query_bracket[0].lower()
            else:
                query_str = queryori.lower()
            return getattr(self,query_str) if hasattr(self,query_str) else None
        else:
            return None


if __name__ == "__main__":
    pass


# from types import SimpleNamespace
# from crown import Model
# from attr import field
from .common import *
from .field import *
from functools import reduce
import operator

class QueryCompiler(object):
    field_map = {
        'int': 'INT',
        'smallint': 'SMALLINT',
        'tinyint': 'TINYINT',
        'bigint': 'BIGINT',
        'float': 'FLOAT',
        'double': 'DOUBLE',
        'nchar': 'NCHAR',
        'binary': 'BINARY',
        'datetime': 'TIMESTAMP',
        'bool': 'BOOL',
        'primary_key': 'TIMESTAMP',
    }

    op_map = {
        OP_EQ: '=',
        OP_LT: '<',
        OP_LTE: '<=',
        OP_GT: '>',
        OP_GTE: '>=',
        OP_NE: '!=',
        OP_IN: 'IN',
        OP_IS: 'IS',
        OP_LIKE: 'LIKE',
        OP_ILIKE: 'ILIKE',
        OP_ADD: '+',
        OP_SUB: '-',
        OP_MUL: '*',
        OP_DIV: '/',
        OP_XOR: '^',
        OP_AND: 'AND',
        OP_OR: 'OR',
    }

    def __init__(self, quote_char='"', interpolation='?', field_overrides=None,
                 op_overrides=None):
        self.quote_char = quote_char
        self.interpolation = interpolation
        self._field_map = dict_update(self.field_map, field_overrides or {})
        self._op_map = dict_update(self.op_map, op_overrides or {})

    def quote(self, s):
        return ''.join((self.quote_char, s, self.quote_char))

    def get_field(self, f):
        return self._field_map[f]

    def get_op(self, q):
        return self._op_map[q]

    def _max_alias(self, am):
        max_alias = 0
        if am:
            for a in am.values():
                i = int(a.lstrip('t'))
                if i > max_alias:
                    max_alias = i
        return max_alias + 1

    def parse_expr(self, expr, alias_map=None):
        s = self.interpolation
        p = [expr]
        if isinstance(expr, Expr):
            lhs, lparams = self.parse_expr(expr.lhs, alias_map)
            rhs, rparams = self.parse_expr(expr.rhs, alias_map)
            s = '(%s %s %s)' % (lhs, self.get_op(expr.op), rhs)
            p = lparams + rparams
        elif isinstance(expr, Field):
            s = expr.db_column
            if alias_map and expr.model_class in alias_map:
                s = '.'.join((alias_map[expr.model_class], s))
            p = []
        elif isinstance(expr, Func):
            p = []
            exprs = []
            for param in expr.params:
                parsed, params = self.parse_expr(param, alias_map)
                exprs.append(parsed)
                p.extend(params)
            
            s = '%s(%s)' % (expr.name, ', '.join(exprs))
        elif isinstance(expr, Param):
            s = self.interpolation
            p = [expr.data]
        elif isinstance(expr, Ordering):
            s, p = self.parse_expr(expr.param, alias_map)
            s += ' ASC' if expr.asc else ' DESC'
        elif isinstance(expr, R):
            s = expr.value
            p = []
        elif isinstance(expr, SelectQuery):
            max_alias = self._max_alias(alias_map)
            clone = expr.clone()
            if not expr._explicit_selection:
                clone._select = (clone.model_class._meta.primary_key,)
            subselect, p = self.parse_select_query(clone, max_alias, alias_map)
            s = '(%s)' % subselect
        elif isinstance(expr, (list, tuple)):
            exprs = []
            p = []
            for i in expr:
                e, v = self.parse_expr(i, alias_map)
                exprs.append(e)
                p.extend(v)
            s = '(%s)' % ','.join(exprs)
        # elif isinstance(expr, Model):
        #     s = self.interpolation
        #     p = [expr.get_id()]

        if isinstance(expr, Leaf):
            if expr.negated:
                s = 'NOT %s' % s
            if expr._alias:
                s = ' '.join((s, 'AS', expr._alias))

        return s, p

    def parse_query_node(self, qnode, alias_map):
        if qnode is not None:
            return self.parse_expr(qnode, alias_map)
        return '', []
    def parse_expr_list(self, s, alias_map):
        parsed = []
        data = []
        for expr in s:
            expr_str, vars = self.parse_expr(expr, alias_map)
            parsed.append(expr_str)
            data.extend(vars)
        return ', '.join(parsed), data


    def parse_select_query(self, query, start=1, alias_map=None):
        model = query.model_class
        db = model._meta.database
        parts = ['SELECT']
        params = []

        if query._distinct:
            parts.append(query._distinct)
        selection = query._select
        select, s_params = self.parse_expr_list(selection, alias_map)

        parts.append(select)
        params.extend(s_params)

        parts.append('FROM %s ' % (self.quote(model._meta.db_table),))

        # joins = self.parse_joins(query._joins, query.model_class, alias_map)
        # if joins:
        #     parts.append(' '.join(joins))

        where, w_params = self.parse_query_node(query._where, alias_map)
        if where:
            parts.append('WHERE %s' % where)
            params.extend(w_params)
        if query._interval:
            if query._interval_offset:
                parts.append('INTERVAL(%s,%s)' % (query._interval,query._interval_offset))
            else:
                parts.append('INTERVAL(%s)' % (query._interval,))
            parts.append('FILL(%s)' % (query._fill,))

        if query._group_by:
            group_by, g_params = self.parse_expr_list(query._group_by, alias_map)
            parts.append('GROUP BY %s' % group_by)
            params.extend(g_params)

        if query._order_by:
            order_by, _ = self.parse_expr_list(query._order_by, alias_map)
            parts.append('ORDER BY %s' % order_by)

        if query._limit or (query._offset and not db.empty_limit):
            limit = query._limit or -1
            parts.append('LIMIT %s' % limit)
        if query._offset:
            parts.append('OFFSET %s' % query._offset)

        return ' '.join(parts), params

    def _parse_field_dictionary(self, d):
        sets, params = [], []
        for item in d:
            field = item.get('obj')
            expr = item.get('value')
            field_str, _ = self.parse_expr(field)
            val_str, val_params = self.parse_expr(expr)
            val_params = [field.db_value(vp) for vp in val_params]
            sets.append((field_str, val_str))
            # sets.append((field_str, val_params[0]))
            params.extend(val_params)
        return sets, params

    def parse_insert_query(self, query):
        model = query.model_class

        parts = ['INSERT INTO %s' % self.quote(model._meta.db_table)]
        sets, params = self._parse_field_dictionary(query._insert)

        parts.append('(%s)' % ', '.join(s[0] for s in sets))
        parts.append('VALUES (%s)' % ', '.join(s[1] for s in sets))

        return ' '.join(parts), params
    def parse_createsontable_query(self,query,safe=True):
        model = query.model_class
        parts = ['CREATE TABLE']
        if safe:
            parts.append('IF NOT EXISTS')
        parts.append('%s USING %s TAGS ' % (query.table_name,model._meta.db_table))
        sets, params = self._parse_field_dictionary(query._tags)
        parts.append('(%s)' % ', '.join(s[1] for s in sets))
        return ' '.join(parts), params

    def field_sql(self, field):
        attrs = field.attributes
        attrs['column_type'] = self.get_field(field.get_db_field())
        template = field.template
        parts = [self.quote(field.db_column), template]
        return ' '.join(p % attrs for p in parts)

    def parse_create_table(self, model_class, safe=False):
        parts = ['CREATE TABLE']
        if safe:
            parts.append('IF NOT EXISTS')
        parts.append(self.quote(model_class._meta.db_table))
        columns = ', '.join(self.field_sql(f) for f in model_class._meta.get_fields())
        parts.append('(%s)' % columns)
        if model_class._tags != None:
            tags = ', '.join(self.field_sql(f) for f in model_class._tags.get_fields())
            parts.append(' TAGS (%s)' % tags)
        return parts

    def parse_create_database(self, database, safe=False,keep= None,comp=None,replica=None,quorum=None,blocks=None):
        parts = ['CREATE DATABASE']
        if safe:
            parts.append('IF NOT EXISTS')
        parts.append(database)
        if keep != None:
            parts.append('KEEP %s' % keep)
        if comp != None:
            parts.append('COMP %s' % comp)
        if replica != None:
            parts.append('REPLICA %s' % replica)
        if quorum != None:
            parts.append('QUORUM %s' % quorum)
        if blocks != None:
            parts.append('BLOCKS %s' % blocks)
        return parts
    def parse_alter_database(self, database,keep= None,comp=None,replica=None,quorum=None,blocks=None):
        parts = ['ALTER DATABASE']
        parts.append(database)
        if keep != None:
            parts.append('KEEP %s' % keep)
        if comp != None:
            parts.append('COMP %s' % comp)
        if replica != None:
            parts.append('REPLICA %s' % replica)
        if quorum != None:
            parts.append('QUORUM %s' % quorum)
        if blocks != None:
            parts.append('BLOCKS %s' % blocks)
        return parts
    def parse_drop_database(self, database, safe=False):
        parts = ['DROP DATABASE']
        if safe:
            parts.append('IF EXISTS')
        parts.append(database)
        return parts
    def create_database(self, database, safe=False,keep= None,comp=None,replica=None,quorum=None,blocks=None):
        return ' '.join(self.parse_create_database(database,safe,keep,comp,replica,quorum,blocks))
    def alter_database(self, database,keep= None,comp=None,replica=None,quorum=None,blocks=None):
        return ' '.join(self.parse_alter_database(database,keep,comp,replica,quorum,blocks))
    def show_database(self, database):
        return 'SHOW DATABASES'
    def show_tables(self, database,super=False):
        if super:
            return 'SHOW %s.STABLES' % database
        else:
            return 'SHOW %s.TABLES' % database
    def drop_database(self, database, safe=False):
        return ' '.join(self.parse_drop_database(database, safe))
    def create_table(self, model_class, safe=False):
        return ' '.join(self.parse_create_table(model_class, safe))
    def describe_table(self,model_class):
        parts = ['DESCRIBE ']
        parts.append(self.quote(model_class._meta.db_table))
        return ' '.join(parts)
    def change_table_tagvalue(self,model_class,name,value):
        parts = ['ALTER TABLE ']
        parts.append(model_class._meta.db_table) #TODO：restful api 对引号的表名支持不好
        # parts.append(self.quote(model_class._meta.db_table))
        if isinstance(value,str):
            parts.append('SET TAG "%s" = "%s"' % (name,value))
        else:
            parts.append('SET TAG "%s" = %s' % (name,value))
        return ' '.join(parts)
    def add_tag(self,model_class,value):
        parts = ['ALTER TABLE ']
        parts.append(model_class._meta.db_table) #TODO：restful api 对引号的表名支持不好
        # parts.append(self.quote(model_class._meta.db_table))
        if isinstance(value, Field) and not isinstance(value, DateTimeField) and value.db_column:
            parts.append('ADD TAG %s' % (self.field_sql(value)))
        return ' '.join(parts)
    def drop_tag(self,model_class,name):
        parts = ['ALTER TABLE ']
        parts.append(model_class._meta.db_table) #TODO：restful api 对引号的表名支持不好
        # parts.append(self.quote(model_class._meta.db_table))
        parts.append('DROP TAG %s' % (name))
        return ' '.join(parts)
    def change_tag_name(self,model_class,name,newname):
        parts = ['ALTER TABLE ']
        parts.append(model_class._meta.db_table) #TODO：restful api 对引号的表名支持不好
        # parts.append(self.quote(model_class._meta.db_table))
        parts.append('CHANGE TAG %s %s' % (name,newname))
        return ' '.join(parts)
    def drop_table(self, model_class, fail_silently=False, cascade=False):
        parts = ['DROP TABLE']
        if fail_silently:
            parts.append('IF EXISTS')
        parts.append(self.quote(model_class._meta.db_table))
        return ' '.join(parts)

class QueryResultWrapper(list):
    def __init__(self, model, cursor):
        self.model = model
        self.cursor = cursor
        cols = []
        non_cols = []
        for i in range(len(self.cursor.head)):
            col = self.cursor.head[i]
            if col in model._meta.columns:
                cols.append((i, model._meta.columns[col]))
            else:
                non_cols.append((i, col))
        self._cols = cols
        self._non_cols = non_cols
        super(QueryResultWrapper, self).__init__([self.simple_mode(row) for row in cursor.data])


    def simple_mode(self, row):
        instance = self.model()
        for i, f in self._cols:
            setattr(instance, f.name, f.python_value(row[i]))
        for i, f in self._non_cols:
            setattr(instance, f, row[i])
        return instance
class Query(object):
    require_commit = True

    def __init__(self, model_class):
        self.model_class = model_class
        self.database = model_class._meta.database

        self._dirty = True
        self._query_ctx = model_class
        # self._joins = {self.model_class: []} # adjacency graph
        self._where = None

    def clone(self):
        query = type(self)(self.model_class)
        if self._where is not None:
            query._where = self._where.clone()
        # query._joins = self.clone_joins()
        query._query_ctx = self._query_ctx
        return query

    @returns_clone
    def where(self, *q_or_node):
        if self._where is None:
            self._where = reduce(operator.and_, q_or_node)
        else:
            for piece in q_or_node:
                self._where &= piece
    def sql(self, compiler):
        raise NotImplementedError()

    def execute(self):
        raise NotImplementedError
class SelectQuery(Query):
    require_commit = False

    def __init__(self, model_class, *selection):
        self._explicit_selection = len(selection) > 0
        all_selection=model_class._meta.get_fields()
        if model_class._tags:
            all_selection.extend(model_class._tags.get_fields())
        self._select = self._model_shorthand(self.strToField(selection , all_selection))
        self._group_by = None
        self._order_by = None
        self._limit = None
        self._offset = None
        self._qr = None
        self._interval= None
        self._interval_offset= None
        self._fill = None
        self._distinct = None
        super(SelectQuery, self).__init__(model_class)
    
    def strToField(self,selection,all_selection):
        if not selection:
            return all_selection
        res = []
        for item in selection:
            if isinstance(item,str):
                for field in all_selection:
                    if item.capitalize() == field.verbose_name or item == field.db_column:
                        item = field
                        break
            res.append(item)
        return res

    def clone(self):
        query = super(SelectQuery, self).clone()
        query._explicit_selection = self._explicit_selection
        query._select = list(self._select)
        if self._group_by is not None:
            query._group_by = list(self._group_by)
        if self._order_by is not None:
            query._order_by = list(self._order_by)
        query._limit = self._limit
        query._offset = self._offset
        query._distinct = self._distinct
        return query

    def _model_shorthand(self, args):
        accum = []
        for arg in args:
            if isinstance(arg, Leaf):
                accum.append(arg)
        return accum
    
    @returns_clone
    def distinct(self):
        self._distinct = 'DISTINCT'

    @returns_clone
    def group_by(self, *args):
        self._group_by = self._model_shorthand(args)

    @returns_clone
    def order_by(self, *args):
        self._order_by = list(args)
    
    @returns_clone
    def desc(self):
        self._order_by = list([self.model_class._meta.primary_key.desc()])

    @returns_clone
    def asc(self):
        self._order_by = list([self.model_class._meta.primary_key.asc()])

    @returns_clone
    def limit(self, lim):
        self._limit = lim

    @returns_clone
    def offset(self, off):
        self._offset = off

    @returns_clone
    def interval(self, interval_value, fill='NONE',offset = None,):
        self._interval= interval_value
        self._interval_offset= offset
        if fill in ['NONE','PREV','NULL','LINEAR']:
            self._fill = fill #{NONE | | PREV | NULL | LINEAR}
        else:
            self._fill = "VALUE, %s" % str(fill) #VALUE

    @returns_clone
    def paginate(self, page, page_size=20):
        if page > 0:
            page -= 1
        self._limit = page_size
        self._offset = page * page_size

    def count(self,field=None):
        clone = self.order_by()
        clone._limit = clone._offset = None
        # TODO: 分组的情况下如何统计
        if self._group_by:
            clone._group_by = None
        clone._select = [fn.Count(field if field else clone.model_class._meta.primary_key)]
        # if self._group_by:
        #     res = clone.execute()
        #     if len(res) > 0:
        #         return res
        #     else:
        #         return None
        # else:
        res = clone.database.execute(clone)
        if res and res[0]:
            return res[0][0]
        else:
            return 0
    def avg(self,*fields):
        clone = self.order_by()
        clone._limit = clone._offset = None
        # TODO: 分组的情况下如何统计
        if self._group_by:
            clone._group_by = None
        clone._select = [out_alias(field,fn.AVG) for field in fields]
        res = clone.execute()
        if len(res) > 0:
            return res[0]
        else:
            return None
    
    def twa(self,*fields):
        clone = self.order_by()
        if clone._where is None:
            return None
        clone._limit = clone._offset = None
        # TODO: 分组的情况下如何统计
        if self._group_by:
            clone._group_by = None
        clone._select = [out_alias(field,fn.TWA) for field in fields]
        res = clone.execute()
        if len(res) > 0:
            return res[0]
        else:
            return None
    def sum(self,*fields):
        clone = self.order_by()
        clone._limit = clone._offset = None
        # TODO: 分组的情况下如何统计
        if self._group_by:
            clone._group_by = None
        clone._select = [out_alias(field,fn.SUM) for field in fields]
        res = clone.execute()
        if len(res) > 0:
            return res[0]
        else:
            return None
    def stddev(self,*fields):
        clone = self.order_by()
        clone._limit = clone._offset = None
        # TODO: 分组的情况下如何统计
        if self._group_by:
            clone._group_by = None
        clone._select = [out_alias(field,fn.STDDEV) for field in fields]
        res = clone.execute()
        if len(res) > 0:
            return res[0]
        else:
            return None
    def min(self,*fields):
        clone = self.order_by()
        clone._limit = clone._offset = None
        # TODO: 分组的情况下如何统计
        if self._group_by:
            clone._group_by = None
        clone._select = [out_alias(field,fn.MIN) for field in fields]
        res = clone.execute()
        if len(res) > 0:
            return res[0]
        else:
            return None
    def max(self,*fields):
        clone = self.order_by()
        clone._limit = clone._offset = None
        # TODO: 分组的情况下如何统计
        if self._group_by:
            clone._group_by = None
        clone._select = [out_alias(field,fn.MAX) for field in fields]
        res = clone.execute()
        if len(res) > 0:
            return res[0]
        else:
            return None
    def first(self,*fields):
        clone = self.order_by()
        clone._limit = clone._offset = None
        # TODO: 分组的情况下如何统计
        if self._group_by:
            clone._group_by = None
        clone._select = [out_alias(field,fn.FIRST) for field in fields]
        res = clone.execute()
        if len(res) > 0:
            return res[0]
        else:
            return None
    def last(self,*fields):
        clone = self.order_by()
        clone._limit = clone._offset = None
        # TODO: 分组的情况下如何统计
        if self._group_by:
            clone._group_by = None
        clone._select = [out_alias(field,fn.LAST) for field in fields]
        res = clone.execute()
        if len(res) > 0:
            return res[0]
        else:
            return None
    def last_row(self,*fields):
        clone = self.order_by()
        if clone._where is not None:
            raise Exception('last_row not allow where clause')
        clone._limit = clone._offset = clone._where = None
        # TODO: 分组的情况下如何统计
        if self._group_by:
            clone._group_by = None
        clone._select = [out_alias(field,fn.LAST_ROW) for field in fields]
        res = clone.execute()
        if len(res) > 0:
            return res[0]
        else:
            return None
    def spread(self,*fields):
        clone = self.order_by()
        clone._limit = clone._offset = None
        # TODO: 分组的情况下如何统计
        if self._group_by:
            clone._group_by = None
        clone._select = [out_alias(field,fn.SPREAD) for field in fields]
        res = clone.execute()
        if len(res) > 0:
            return res[0]
        else:
            return None
    # tdengine目前只支持一列，为以后支持多列准备的函数
    def diff(self,*fields):
        clone = self.order_by()
        clone._limit = clone._offset = None
        # TODO: 分组的情况下如何统计
        if self._group_by:
            clone._group_by = None
        clone._select = [out_alias(field,fn.DIFF) for field in fields]
        res = clone.execute()
        return res
    def top(self,field,top=1,alias=None):
        clone = self.order_by()
        clone._limit = clone._offset = None
        # TODO: 分组的情况下如何统计
        if self._group_by:
            clone._group_by = None
        if alias:
            clone._select = [fn.TOP(field,top).alias(alias)]
        else:
            clone._select = [fn.TOP(field,top)]
        res = clone.execute()
        return res
    def bottom(self,field,bottom=1,alias=None):
        clone = self.order_by()
        clone._limit = clone._offset = None
        # TODO: 分组的情况下如何统计
        if self._group_by:
            clone._group_by = None
        if alias:
            clone._select = [fn.BOTTOM(field,bottom).alias(alias)]
        else:
            clone._select = [fn.BOTTOM(field,bottom)]
        
        res = clone.execute()
        return res
    def apercentile(self,*fields):
        clone = self.order_by()
        clone._limit = clone._offset = None
        # TODO: 分组的情况下如何统计
        if self._group_by:
            clone._group_by = None
        clone._select = [out_alias_tuple_field(field,3,fn.APERCENTILE) for field in fields]
        res = clone.execute()
        if len(res) > 0:
            return res[0]
        else:
            return None
    def percentile(self,*fields):
        clone = self.order_by()
        clone._limit = clone._offset = None
        # TODO: 分组的情况下如何统计
        if self._group_by:
            clone._group_by = None
        clone._select = [out_alias_tuple_field(field,3,fn.PERCENTILE) for field in fields]
        res = clone.execute()
        if len(res) > 0:
            return res[0]
        else:
            return None
    def leastsquares(self,*fields):
        clone = self.order_by()
        clone._limit = clone._offset = None
        # TODO: 分组的情况下如何统计
        if self._group_by:
            clone._group_by = None
        clone._select = [out_alias_tuple_field(field,4,fn.LEASTSQUARES) for field in fields]
        res = clone.execute()
        if len(res) > 0:
            return res[0]
        else:
            return None

    def exists(self):
        clone = self.paginate(1, 1)
        res = clone.execute()
        if len(res) >0:
            return True
        else:
            return False

    def one_raw(self):
        clone = self.paginate(1, 1)
        res = clone.database.execute(clone)
        if len(res) >0:
            return res[0]
        else:
            return None
    
    def one(self):
        clone = self.paginate(1, 1)
        res = clone.execute()
        if len(res) >0:
            return res[0]
        else:
            return None
       
    def sql(self, compiler):
        return compiler.parse_select_query(self)

    def execute(self):
        if self._dirty or not self._qr:
            self._qr = QueryResultWrapper(self.model_class, self.database.execute(self))
            self._dirty = False
            return self._qr
        else:
            return self._qr
    def all_raw(self):
        return self.database.execute(self)
    def all(self):
        return self.execute()
class InsertQuery(Query):
    def __init__(self, model_class, insert=None):
        # mm = model_class._meta
        # query = dict((mm.fields[f], v) for f, v in mm.get_default_dict().items())
        # query.update(insert)
        self._insert = insert
        super(InsertQuery, self).__init__(model_class)

    def clone(self):
        query = super(InsertQuery, self).clone()
        query._insert = list(self._insert)
        return query

    where = not_allowed('where clause')

    def sql(self, compiler):
        return compiler.parse_insert_query(self)

    def execute(self):
        result = self.database.execute(self)
        return result
class CreateSonTableQuery(Query):
    def __init__(self, model_class, values=None,table_name=None):
        # mm = model_class._tags
        # query = dict((mm.fields[f], v) for f, v in mm.get_default_dict().items())
        # query.update(values)
        self._tags = values
        self.table_name = "%s.%s" % (model_class._meta.database.database,table_name)
        super(CreateSonTableQuery, self).__init__(model_class)

    where = not_allowed('where clause')

    def clone(self):
        query = super(CreateSonTableQuery, self).clone()
        query._tags = list(self._tags)
        return query

    def sql(self, compiler):
        return compiler.parse_createsontable_query(self)

    def execute(self):
        result = self.database.execute(self)
        return result


from .crown import Model,SuperModel
from .common import DoesNotExist,EmptyResultException,logger
from .database import TdEngineDatabase
from .field import fn,IntegerField,BigIntegerField,SmallIntegerField,\
                   TinyIntegerField,FloatField,DoubleField,NCharField,\
                   BinaryField,DateTimeField,PrimaryKeyField,BooleanField

                   
__version__ = "0.0.1"


import logging
from collections import namedtuple
OP_AND = 0
OP_OR = 1

OP_ADD = 10
OP_SUB = 11
OP_MUL = 12
OP_DIV = 13
OP_AND = 14
OP_OR = 15
OP_XOR = 16
OP_USER = 19

OP_EQ = 20
OP_LT = 21
OP_LTE = 22
OP_GT = 23
OP_GTE = 24
OP_NE = 25
OP_IN = 26
OP_IS = 27
OP_LIKE = 28
OP_ILIKE = 29

# JOIN_INNER = 1
# JOIN_LEFT_OUTER = 2
# JOIN_FULL = 3

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger('crown')



Ordering = namedtuple('Ordering', ('param', 'asc'))
R = namedtuple('R', ('value',))
# Join = namedtuple('Join', ('model_class', 'join_type', 'on'))
def out_alias(field,fun):
    if field._alias:
        name = field._alias
        field._alias = None
        return fun(field).alias(name)
    else:
        return fun(field)

def out_alias_tuple_field(field,n,fun):
    if not hasattr(field, '__len__'):
        raise Exception('field is not a tuple or list')
    if len(field) < n-1:
        raise Exception('field param less than %i' % (n-1))
    if len(field) == n:
        name = field[n-1]
        return fun(*field[:n-1]).alias(name)
    else:
        return fun(*field[:n-1])

def dict_update(orig, extra):
    new = {}
    new.update(orig)
    new.update(extra)
    return new

def returns_clone(func):
    def inner(self, *args, **kwargs):
        clone = self.clone()
        func(clone, *args, **kwargs)
        return clone
    inner.call_local = func
    return inner

def not_allowed(fn):
    def inner(self, *args, **kwargs):
        raise NotImplementedError('%s is not allowed on %s instances' % (
            fn, type(self).__name__,
        ))
    return inner

class DoesNotExist(Exception):
    pass

# doing an IN on empty set
class EmptyResultException(Exception):
    pass


from setuptools import setup
from os import path

DIR = path.dirname(path.abspath(__file__))
# INSTALL_PACKAGES = open(path.join(DIR, 'requirements.txt')).read().splitlines()

with open(path.join(DIR, 'README.rst'),encoding='utf-8') as f:
    README = f.read()

setup(
    name='crown',
    packages=['crown'],
    description="crown is a simple and small ORM for Time Series Database (TSDB) tdengine(taos), making it easy to learn and intuitive to use.",
    long_description=README,
    long_description_content_type='text/markdown',
    install_requires=[
        'requests>=2.23.0'
    ],
    version='0.1.7',
    url='https://github.com/machine-w/crown',
    author='machine-w',
    author_email='steve2008.ma@gmail.com',
    keywords=['orm','taos', 'TDengine', 'TSDB','Time Series Database','connector','python'],
    tests_require=[
        'pytest',
        'pytest-faker',
        'pytest-sugar',
        'numpy',
        'pandas'
    ],
    package_data={
        # include json and txt files
        '': ['*.rst','*.txt'],
    },
    include_package_data=True,
    python_requires='>=3'
)